import pandas as pd
import numpy as np
from pathlib import Path
from datetime import datetime, date, timedelta, time
import json
from dataclasses import dataclass
from enum import Enum


def start_of_day(day: date, tzinfo=None) -> datetime:
    return datetime.combine(day, time(tzinfo=tzinfo))


def strpdelta(s: str, fmt: str = "%HH%MM") -> timedelta:
    return datetime.strptime(s, fmt) - datetime.strptime("00H00M", "%HH%MM")


def time_to_timedelta(t: time) -> timedelta:
    return datetime.combine(date.today(), t) - start_of_day(date.today())


def to_timedelta(t, fmt=None) -> timedelta:
    if isinstance(t, timedelta):
        return t
    elif isinstance(t, datetime):
        return time_to_timedelta(t.time())
    if fmt is None:
        return strpdelta(str(t))
    else:
        return strpdelta(str(t), fmt)


def to_date(d) -> date:
    if isinstance(d, date):
        return d
    elif isinstance(d, (datetime, pd.Timestamp)):
        return d.date()
    try:
        return datetime.strptime(str(d), "%Y%m%d").date()
    except:
        return datetime.strptime(str(d), "%Y-%m-%d").date()


class TSessKind(Enum):
    DAY = 1
    NIGHT = 2


@dataclass
class TSess:
    start: timedelta
    end: timedelta
    kind: TSessKind


@dataclass
class DateTSess:
    start: datetime
    end: datetime
    kind: TSessKind


def make_session(s, e, is_day: bool = True) -> TSess:
    kind = TSessKind.DAY if is_day else TSessKind.NIGHT
    return TSess(start=to_timedelta(s), end=to_timedelta(e), kind=kind)


ZCE_TZ_STR = "Asia/Shanghai"


def to_date_session(s: TSess, d: date) -> DateTSess:
    def to_dt(d: date, delta: timedelta) -> datetime:
        return (pd.Timestamp(d, tz=ZCE_TZ_STR) + delta).to_pydatetime()

    s = DateTSess(to_dt(d, s.start), to_dt(d, s.end), s.kind)
    assert s.start < s.end
    return s


ZCE_TRADING_SESSION = [
    make_session("21H00M", "23H00M", is_day=False),
    make_session("09H00M", "10H15M"),
    make_session("10H30M", "11H30M"),
    make_session("13H30M", "15H00M"),
]

HOLIDAY_WEIGHT = 0.2

WEEKEND_WEIGHTS_BY_BOARD = {"SR": 1.5542}


DFLT_HOLIDAY_FILENAME = "cn_holidays.json"


def load_holiday_list_json_file(path: Path) -> list[date]:
    with open(Path(path), "r") as f:
        return [to_date(dint) for dint in json.load(f)["holidays"]]


class DayKind(Enum):
    NORMAL_BDAY = 1  # has night session
    MONDAY_BDAY_WITH_WEEKEND_WEIGHT = 2  # has night session
    POST_HOLIDAY_BDAY = 3  # has NO night session
    WEEKEND = 4
    HOLIDAY = 5


def is_bday(kind: DayKind) -> bool:
    return kind in (
        DayKind.NORMAL_BDAY,
        DayKind.MONDAY_BDAY_WITH_WEEKEND_WEIGHT,
        DayKind.POST_HOLIDAY_BDAY,
    )


def has_night_session(k: DayKind) -> bool:
    return k in (DayKind.NORMAL_BDAY, DayKind.MONDAY_BDAY_WITH_WEEKEND_WEIGHT)


DecayProfiles = dict[DayKind, pd.DataFrame]


def load_ttm_decay_csvs_from_dir(dir: Path, board: str) -> DecayProfiles:
    decay_profiles = {}
    for k, p in TTM_DECAY_CSV_FILENAME.items():
        fpath = Path(dir) / p.format(board)
        if not fpath.is_file():
            raise ValueError(f"missing file {fpath}")
        s = pd.read_csv(fpath)
        s.timedelta = pd.to_timedelta(s.timedelta)
        decay_profiles[k] = s
    return decay_profiles


@dataclass
class WeightedCalendar:
    board: str
    wdf: pd.DataFrame
    wsum_by_year: dict[int, float]

    def dump_wdf(self, p: Path):
        self.wdf.to_csv(p, index=True, index_label="date")

    def get_trading_session(self, d: date) -> list[DateTSess]:
        d = to_date(d)
        drow = self._select_wdf_by_date(d)
        if not is_bday(drow.kind):
            return []
        NIGHTS = ZCE_TRADING_SESSION[0]
        assert NIGHTS.kind == TSessKind.NIGHT
        slist = []
        if has_night_session(drow.kind):
            prev_bday = self.get_prev_bday(d)
            slist.append(to_date_session(NIGHTS, prev_bday))
        for s in ZCE_TRADING_SESSION[1:]:
            assert s.kind != TSessKind.NIGHT
            slist.append(to_date_session(s, d))
        return slist

    def get_prev_bday(self, d: date) -> date:
        curi = self._get_date_df_loc(d)
        while curi > 0:
            curi -= 1
            if is_bday(self.wdf.iloc[curi].kind):
                return self.wdf.index[curi]
        raise ValueError(f"{d} out of bound!")

    def weight_sum_inclusive(self, start: date, end: date) -> float:
        """return sum of weights, including start/end dates."""
        start, end = to_date(start), to_date(end)
        range = self.wdf.query("index >= @start & index <= @end")
        bound = (self.wdf.index[0].date(), self.wdf.index[-1].date())
        if range.shape[0] == 0 or start < bound[0] or end > bound[1]:
            raise ValueError(
                f"out of bound: [{start}, {end}]. date bounds: [{bound[0]}, {bound[1]}]"
            )
        return range.weight.sum()

    def gen_dtm(
        self, board: str, start: date, end: date, decay_profiles: DecayProfiles
    ):
        """
        if a day is not a trading day:
        1) it is a weekend, monday night session end
        2) it is a holiday.
        if a day is a trading day:
        1) it is a monday that has extra weekend weights.
        2) it is post holiday
        3) otherwise,
        """
        start, end = to_date(start), to_date(end)
        if not is_bday(self._select_wdf_by_date(start).kind):
            raise ValueError(f"start [{start}] shall be a trading day")
        if not is_bday(self._select_wdf_by_date(end).kind):
            raise ValueError(f"end [{end}] shall be a trading day")
        rangedf = self.wdf[
            (self.wdf.index.date >= start) & (self.wdf.index.date <= end)
        ].copy()
        if rangedf.shape[0] == 0:
            raise ValueError(f"empty [{start}, {end}]")
        # this column is inaccurate for non-trading day, since their decay starts at previous close rather than midnight
        # the sod of trading day is the start time of first session.
        rangedf["trading_day_sod_dtm"] = rangedf.weight[::-1].cumsum()
        trading_days = rangedf[rangedf.kind.apply(is_bday)]

        #
        num_bdays_per_yr = self.wsum_by_year[start.year]
        combined_rows = []
        for d, dayrow in trading_days.iterrows():
            d = to_date(d)
            sess_list = self.get_trading_session(d)
            assert len(sess_list) > 0
            if dayrow.kind not in decay_profiles:
                raise ValueError(f"{d} illegal kind")

            sidx = 0
            profile = decay_profiles[dayrow.kind].copy()
            if board not in profile.columns:
                raise ValueError(f"[{board}] and profile mismatch!")
            assert profile[board].iloc[-1] == 1.0
            profile[board] *= dayrow.weight

            def within_session(s: DateTSess, td) -> bool:
                # assume no across-midnight session
                assert s.start.date() == s.end.date()
                std = s.start - start_of_day(s.start.date(), s.start.tzinfo)
                etd = s.end - start_of_day(s.end.date(), s.end.tzinfo)
                return td >= std and td <= etd

            for _, row in profile.iterrows():
                s = sess_list[sidx]
                while not within_session(s, row.timedelta):
                    sidx += 1
                    if sidx == len(sess_list):
                        raise ValueError(
                            f"timedelta {row.timedelta} not in any trading session {sess_list}"
                        )
                    s = sess_list[sidx]

                t: pd.Timestamp = (
                    start_of_day(s.start.date(), s.start.tzinfo) + row.timedelta
                )
                dtm = dayrow.trading_day_sod_dtm - row[board]
                combined_rows.append(
                    {
                        "time": t,
                        "dtm": dtm,
                        "daily_init_dtm": dayrow.trading_day_sod_dtm,
                        "daily_weight": dayrow.weight,
                        "num_bdays_per_yr": num_bdays_per_yr,
                    }
                )
        return pd.DataFrame(combined_rows)

    def has_night_session(self, d: date) -> bool:
        return has_night_session(self._select_wdf_by_date(d).kind)

    def _get_date_df_loc(self, d: date) -> int:
        return self.wdf.index.get_loc(pd.Timestamp(to_date(d)))

    def _select_wdf_by_date(self, d: date) -> pd.Series:
        return self.wdf.iloc[self._get_date_df_loc(d)]


def gen_daily_weights(board: str, holiday_list: list[date]) -> WeightedCalendar:
    if board not in WEEKEND_WEIGHTS_BY_BOARD:
        raise ValueError(f"board [{board}] weekend weight is missing!")
    # use first and last holiday to decide start/end years
    # holiday list shall contain all holidays within [start_year, end_year]
    b, e = holiday_list[0], holiday_list[-1]
    all_dates = pd.DataFrame(
        index=pd.date_range(f"1/1/{b.year}", f"12/31/{e.year}", freq="D")
    )
    all_dates["weekday"] = all_dates.index.dayofweek
    all_dates["is_holiday"] = all_dates.index.isin(holiday_list)
    all_dates["is_weekend"] = all_dates.weekday >= 5
    all_dates["is_non_bday"] = all_dates.is_holiday | all_dates.is_weekend
    all_dates["is_bday"] = ~all_dates.is_non_bday
    # weekend next to holiday shall be considered holiday too.
    all_dates["cnt_consecutive_non_bday"] = all_dates.is_non_bday.groupby(
        all_dates.is_non_bday.ne(all_dates.is_non_bday.shift()).cumsum()
    ).transform("size")
    all_dates["cnt_consecutive_non_bday"].where(all_dates.is_non_bday, 0, inplace=True)
    # holiday_grp: holiday and weekends next to holiday
    # holiday_grp uses holiday weight
    all_dates["is_holiday_grp"] = all_dates.is_holiday | (
        all_dates.cnt_consecutive_non_bday >= 3
    )
    # used to decide if a trading day has night session or not.
    # first date is always 1/1, a holida's yd_is_holiday_grp doesn't matter
    all_dates["yd_is_holiday_grp"] = all_dates.is_holiday_grp.shift(1, fill_value=False)
    # monday that has night session shall use the weekend weight multiplier
    # a monday has night session if it is trading day and the previous weekend is not next to in a holiday
    all_dates["assign_weekend_weight_to_monday"] = (
        (all_dates.weekday == 0)
        & (~all_dates.yd_is_holiday_grp)
        & (~all_dates.is_non_bday)
    )
    all_dates["post_holiday_bday"] = (
        ~all_dates.is_non_bday
    ) & all_dates.yd_is_holiday_grp
    all_dates["kind"] = DayKind.NORMAL_BDAY
    all_dates.loc[all_dates.is_holiday_grp, "kind"] = DayKind.HOLIDAY
    all_dates.loc[all_dates.assign_weekend_weight_to_monday, "kind"] = (
        DayKind.MONDAY_BDAY_WITH_WEEKEND_WEIGHT
    )
    all_dates.loc[all_dates.post_holiday_bday, "kind"] = DayKind.POST_HOLIDAY_BDAY
    all_dates.loc[all_dates.is_weekend & (~all_dates.is_holiday_grp), "kind"] = (
        DayKind.WEEKEND
    )
    all_dates["weight"] = 1.0
    all_dates.loc[all_dates.kind == DayKind.WEEKEND, "weight"] = 0
    all_dates.loc[all_dates.kind == DayKind.HOLIDAY, "weight"] = HOLIDAY_WEIGHT
    all_dates.loc[
        all_dates.kind == DayKind.MONDAY_BDAY_WITH_WEEKEND_WEIGHT, "weight"
    ] = WEEKEND_WEIGHTS_BY_BOARD[board]
    assert np.all(all_dates.kind.apply(is_bday) != all_dates.is_non_bday)
    assert np.all(all_dates.kind.apply(is_bday) == all_dates.is_bday)
    sums = {}
    for i, w in all_dates.groupby(all_dates.index.year).weight.sum().items():
        sums[int(i)] = w
    return WeightedCalendar(
        board=board,
        wdf=all_dates[["weekday", "is_holiday", "weight", "kind"]],
        wsum_by_year=sums,
    )


TTM_DECAY_CSV_FILENAME = {
    DayKind.NORMAL_BDAY: "{}_weekday1to4.csv",
    DayKind.MONDAY_BDAY_WITH_WEEKEND_WEIGHT: "{}_weekday0.csv",
    DayKind.POST_HOLIDAY_BDAY: "{}_postholiday.csv",
}


def tests():
    pass


if __name__ == "__main__":

    from pathlib import Path

    SRCP = Path(__file__).resolve()
    SRCD = SRCP.parent
    import click

    @click.group()
    def cli():
        pass

    @cli.command()
    @click.option(
        "--idir", default=SRCD, help="input dir that contains .csv/holidays.json"
    )
    @click.argument("board")
    @click.argument("start")
    @click.argument("end")
    @click.argument("dst")
    def dumpweights(idir, board, start, end, dst):
        p = load_ttm_decay_csvs_from_dir(Path(idir), board)
        hlist = load_holiday_list_json_file(Path(idir) / DFLT_HOLIDAY_FILENAME)
        gen_daily_weights(board, hlist).dump_wdf(dst)

    @cli.command()
    @click.option(
        "--idir", default=SRCD, help="input dir that contains .csv/holidays.json"
    )
    @click.argument("board")
    @click.argument("date", default=None)
    def gen(idir, board, date):
        pass

    cli()
